const express = require('express');
const { spawn } = require('child_process');
const { spawnSync } = require('child_process');
const crypto = require('crypto');
const redis = require('redis');
const WebSocket = require('ws');
const http = require('http');
const path = require('path');
const bodyParser = require('body-parser');
const start = "a";        // limite de début
const end = "9999999999";     // limite de fin

const {generateSimulatedHash, sendHashToSlave} = require('./utils');

const app = express();

app.use(express.json());
app.use(bodyParser.json());
app.use(bodyParser.urlencoded({extended: true}));

// le client Redis
const client = redis.createClient({
    url: 'redis://redis:6379'
});

client.on('error', (err) => console.log('Erreur Redis Client', err));

// etablissement de la connexion Redis
async function connectRedis() {
    try {
        await client.connect();
    } catch (err) {
        console.error('Erreur de connexion Redis:', err);
        process.exit(1);
    }
}

connectRedis();

// Création du serveur HTTP
const server = http.createServer(app);

// Création du serveur WebSocket en attachant au serveur HTTP
const wss = new WebSocket.Server({server});

let esclaves = new Map();
let esclaveId = 0;

// Tableau pour stocker les tâches en attente
let pendingTasks = [0];

let runningContainerId = []; // Variable pour stocker l'ID du conteneur en cours d'exécution

// Gestionnaire d'ouverture de la connexion WebSocket
wss.on('connection', (ws) => {
    console.log('Connexion WebSocket avec un client établie');
    esclaveId++;
    esclaves.set(esclaveId, ws);
    console.log('size', esclaves.size);

    // Gestion des messages WebSocket
    ws.on('message', async (message) => {
        console.log('Message reçu du client:', message.toString('utf-8'));
        const decodedMessage = message.toString('utf-8');

        // Vérifiez si le message commence par "found "
        if (decodedMessage.startsWith('found ')) {
            // Extrait le hash MD5 de la partie "found " du message
            const md5Hash = decodedMessage.substring(6, 38);
            const text = decodedMessage.substring(39);

            for (let i = pendingTasks.length - 1; i >= 0; i--){
                if (pendingTasks[i] > 0) {
                    pendingTasks[i]--;
                    if (pendingTasks[i] === 0) {
                        const removedContainerId = runningContainerId.pop();
                        /*
                        try {
                            process.kill(removedContainerId, 'SIGKILL');
                            console.log('Processus Docker tué avec succès.');
                        } catch (err) {
                            console.error('Erreur lors de la tentative de tuer le processus Docker :', err.message);
                        }
                        */
                        const result = spawnSync('docker', ['stop', removedContainerId]);

                        if (result.error) {
                            console.error('Erreur lors de l\'arrêt du conteneur Docker :', result.error.message);
                        } else {
                            console.log('Conteneur Docker arrêté avec succès.');
                            if (esclaveId > 0) {
                                esclaves.delete(esclaveId);
                                esclaveId--;
                            }
                        }

                        //spawn('docker', ['stop', removedContainerId]);
                        pendingTasks.splice(i, 1);

                    }
                    break;
                }
            }

            console.log('tache en attente', pendingTasks, "=====================", runningContainerId);

            // Stocke le hash MD5 dans Redis
            await client.set(md5Hash, text);
        }
    });

    // Gestion de la déconnexion
    ws.on('close', () => {
        console.log('Client déconnecté');
        if (esclaveId > 0) {
            esclaves.delete(esclaveId);
            esclaveId--;
        }
    });
});

app.get('/', (req, res) => {
    res.sendFile(path.join(__dirname, 'index.html'));
});

app.post('/generateHash', async (req, res) => {
    const difficulty = req.body.complexity;

    if (!difficulty) {
        return res.status(400).json({error: 'Missing data in the request.'});
    }

    const hash = generateSimulatedHash(difficulty);
    const searchToRedis = await client.get(hash);

    let wsEsclave = esclaves.get(1);

    if (searchToRedis === null) {
        /* wss.clients.forEach((client) => {
            if (client.readyState === WebSocket.OPEN) {
                sendHashToSlave(hash, client, start, end);
            }
        }); */
        wsEsclave.send(hash);
    } else {
        console.log('Le mot de passe a été trouvé dans la base de données Redis' , searchToRedis);
    }

    res.redirect('/');
});

app.post('/generateHashOfPassword', async (req, res) => {
    try {
        const password = req.body.password;

        if (!password) {
            return res.status(400).json({error: 'Missing data in the request.'});
        }

        const hash = crypto.createHash('md5').update(password).digest('hex');

        const searchToRedis = await client.get(hash);

        let wsEsclave = esclaves.get(1);

        if (searchToRedis === null) {
            // Envoi du texte au service esclave via WebSocket
            /*
            wss.clients.forEach((client) => {
                if (client.readyState === WebSocket.OPEN) {
                    sendHashToSlave(hash, client, start, end);
                }
            });
            */
            wsEsclave.send(hash);
        } else {
            console.log('Le mot de passe a été trouvé dans la base de données Redis', searchToRedis);
        }

        res.redirect('/');
    } catch (err) {
        res.status(500).json({error: 'Erreur du serveur'});
    }
});

const PORT = 3000;
server.listen(PORT, () => {
    console.log(`Serveur démarré sur le port ${PORT}`);
});


const socket = new WebSocket('ws://127.0.0.1:3000');

socket.addEventListener('open', (event) => {
    console.log('Connexion WebSocket établie avec le serveur');
});

socket.addEventListener('message', (event) => {
    console.log('Message reçu du serveur:', event.data);
    const message = event.data;

    const wsEsclave = esclaves.get(esclaveId);
    pendingTasks[esclaveId - 2 ]++;

    console.log('tache en attente', pendingTasks, "=====", runningContainerId, "=======", esclaveId);

    sendHashToSlave(message, wsEsclave, start, end);

    if ( pendingTasks[esclaveId - 2 ] >= 2){
        console.log('esclaveId', esclaveId, "esclaves.size", esclaves.size);

        console.log('Lancement d\'un nouveau conteneur Docker...');

        dockerProcess = spawn('docker', ['run', '--network', 'host', '--rm', 'servuc/hash_extractor:latest', './hash_extractor', 's', 'ws://127.0.0.1:3000'], { stdio: 'inherit' });

        // Stockez l'ID du nouveau conteneur en cours d'exécution
        runningContainerId[esclaveId - 2] = dockerProcess.pid;

        pendingTasks.push(0);

        console.log('==================', pendingTasks, "=======", runningContainerId);
    }
});

module.exports = { start, end, esclaves, esclaveId, pendingTasks, socket };
